/* memspd.c
xref:
input:
output:
does: A quick test of memory speed
      Compare time to sum a matrix by rows and by columns
*/
/************************************************************/
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <float.h>
#include <limits.h>

#define DEBUG 0

#define VERSION "Memory Speed, v 1.0\n"
#define AUTHOR  "(c) Bahjat Qaqish, 2003\n"

typedef unsigned char  UCHAR;
typedef unsigned short USHORT;
typedef unsigned int   UINT;
typedef unsigned long  ULONG;
#define HALFMAX (ULONG_MAX/2)

#define SEED 12345
#define MIN_TIME 0.5
#define N     1024

typedef double VECTOR [N];
typedef VECTOR MATRIX [N];

double matsum_byrow (MATRIX a, int m, int n);
double matsum_bycol (MATRIX a, int m, int n);

void FillRand (MATRIX x, int m, int n);
double myrandom (void);
double get_time (double f (MATRIX,int,int), ULONG  niter);
ULONG  get_niter (double f (MATRIX,int,int), double min_time);

MATRIX global_a;
int    global_junk = 0;

/************************************************************/
int main(int argc, char **argv)
{
  double t_r, t_c;
  ULONG niter;

  printf ("%s", VERSION);
  printf ("%s", AUTHOR );
  printf ("\n");
  printf("Summing a %dx%d matrix by row and by column\n",N,N);
  printf("Matrix size is %d bytes\n", N * N * sizeof global_a[0][0]);

  niter = get_niter (matsum_byrow, MIN_TIME);

  t_r  = get_time(matsum_byrow, niter);
  t_c  = get_time(matsum_bycol, niter);

  printf("# iterations: %d\n", niter );
  printf("by_row:   %g seconds per iteration\n", t_r / niter );
  printf("by_col:   %g seconds per iteration\n", t_c / niter );
  printf("by_col/by_row ratio:  %g\n", t_c/t_r);
  printf("by_row/by_col ratio:  %g\n", t_r/t_c);
  printf("by_row: %g nanoseconds per element\n", 1e9*t_r/niter/N/N);
  printf("by_col: %g nanoseconds per element\n", 1e9*t_c/niter/N/N);

  return global_junk;
}

/************************************************************/
double matsum_byrow(MATRIX a, int m, int n)
/* return matrix sum, m = #rows, n = #cols, 0-based indexing */
{
  int i, j;
  double s=0;

  for (i=0; i < m; ++i)
    for (j=0; j < n; ++j)   s += a[i][j];

  return s;
}
/************************************************************/
double matsum_bycol(MATRIX a, int m, int n)
/* return matrix sum, m = #rows, n = #cols, 0-based indexing */
{
  int i, j;
  double s=0;

  for (j=0; j < n; ++j)
    for (i=0; i < m; ++i) s += a[i][j];

  return s;
}
/************************************************************/
void FillRand (MATRIX x, int m, int n)
  /* fill x with random #'s  */
{
  int i, j;

  for (i=0; i < m; ++i)
    for (j=0; j < n; ++j)
      x[i][j] = myrandom();
}
/************************************************************/
double myrandom (void)
  /* generate the next pseudo-random number */
  /* primitive, but fine for this */
{
  const int m = 134456;
  const int a = 8121;
  const int c = 28411;
  static int k = SEED;

  k = (k*a+c) % m;  /*  0 <= k <= m-1 */
  return( k / (double) m );
}

/************************************************************/
double get_time (double f (MATRIX,int,int), ULONG niter)
{
  double t;
  ULONG i;

  FillRand(global_a, N, N);
  t = clock();
  for (i=0; i != niter; ++i)
    global_junk += (0 == f(global_a, N, N));
  t = (clock()-t) / CLOCKS_PER_SEC;
  return t;
}

/************************************************************/
ULONG  get_niter (double f (MATRIX,int,int), double min_time)
/* returns min # iterations requiring  >= min_time seconds */
{
  ULONG n = 1;
  double t;

  if (min_time <= 0) return (0);

  do {
    if (n > HALFMAX) {printf("time_func: n too large\n"); exit(1);}
    n += n;
    t = get_time (f, n);
  } while (t < min_time);

  return n;
}
