/* pp2.c
xref:
input:
output: stdout
does:   - obtain a general performance profile
        - time int+int, dbl+dbl, log(), exp(), etc
        - the code tries to prevent the compiler from removing the loops
*/
/************************************************************/
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <float.h>
#include <limits.h>

#define DEBUG 0

#define VERSION "Performance Profile, v 2.1\n"
#define AUTHOR  "(c) Bahjat Qaqish, 2005\n"
#define COMNT1  "All times that follow are compensated for 'nop'\n"

#define FACTOR 8
#define MIN_TIME 0.5

double nop_time = 0;  /* nop time in nanoseconds */

typedef unsigned char  UCHAR;
typedef unsigned short USHORT;
typedef unsigned int   UINT;
typedef unsigned long  ULONG;
#define HALFMAX (ULONG_MAX/2)

double nop_t    (ULONG n);
double ipp_t    (ULONG n);
double i_ipi_t  (ULONG n);
double i_iti_t  (ULONG n);

double fpp_t    (ULONG n);
double f_fpf_t  (ULONG n);
double f_ftf_t  (ULONG n);

double dpp_t    (ULONG n);
double d_dpd_t  (ULONG n);
double d_dtd_t  (ULONG n);

double sqrt_t (ULONG n);

double sin_t  (ULONG n);
double cos_t  (ULONG n);

double log_t  (ULONG n);
double exp_t  (ULONG n);

double  time_func (double f(ULONG), const char *name);
void show (const char* s, double t, ULONG n);

double sprintf_t (ULONG n);
/************************************************************/

int main(int argc, char **argv)
{
  double t;

  printf ("%s", VERSION);
  printf ("%s", AUTHOR );
  printf ("\n");

/*
  nop_time = time_func (nop_t, "nop (uncomp)");
  printf ("%s", COMNT1 );
  time_func (nop_t, "nop");
*/
  time_func (ipp_t, "int++");
  time_func (i_ipi_t, "int=int+int");
  time_func (i_iti_t, "int=int*int");

  time_func (fpp_t, "flt++");
  time_func (f_fpf_t, "flt=flt+flt");
  time_func (f_ftf_t, "flt=flt*flt");

  time_func (dpp_t, "dbl++");
  time_func (d_dpd_t, "dbl=dbl+dbl");
  time_func (d_dtd_t, "dbl=dbl*dbl");

  time_func (sqrt_t, "dbl=sqrt(dbl)");
  time_func (log_t, "dbl=log(dbl)");
  time_func (exp_t, "dbl=exp(dbl)");
  time_func (sin_t, "dbl=sin(dbl)");
  time_func (cos_t, "dbl=cos(dbl)");

  time_func (sprintf_t, "sprintf");

  return  0;
}


/************************************************************/
double  time_func (double f(ULONG), const char *name)
/* time a function and show results */
/* returns the time in nanoseconds per operation */
{
  ULONG  n=1;
  double t;

  do {
    if (n > HALFMAX) {printf("time_func: n too large\n"); exit(1);}
    n += n;
    t = f(n);
  } while (t < MIN_TIME);

  t = (1.0e9/FACTOR)*(t/n) - nop_time;
  show(name, t, n);

  return t;

}
/************************************************************/
void show (const char* name, double t, ULONG n)
{
  printf ("%-15s %8g nanoseconds\n",  name, t);
}
/************************************************************/

double nop_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i, j;
  double t;

#if DEBUG
  printf("nop_t: n = %g \n", (double)n);
#endif

  t  = clock();
  for (j=0; j != FACTOR; ++j)
      for (i=0; i != n; ++i) { }
  t = (clock() - t) / CLOCKS_PER_SEC;
  return  t + (i > n); /* (i > n) is to trick the compiler */
}
/************************************************************/

double ipp_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  int x1, x2, x3, x4, x5, x6, x7, x8;

#if DEBUG
  printf("ipp_t: n = %g \n", (double)n);
#endif

  x1=x2=x3=x4=x5=x6=x7=x8= INT_MIN;

  t  = clock();
  for (i=0; i != n; ++i) {
    x1++ ;
    x2++ ;
    x3++ ;
    x4++ ;
    x5++ ;
    x6++ ;
    x7++ ;
    x8++ ;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = abs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double i_ipi_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  int x1, x2, x3, x4, x5, x6, x7, x8;

#if DEBUG
  printf("i_ipi_t: n = %g \n", (double)n);
#endif

  x1=x2=x3=x4=x5=x6=x7=x8 = 1;

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 + x2;
    x3= x4 + x4;
    x5= x6 + x6;
    x7= x8 + x8;
    x2= x1 + x1;
    x4= x3 + x3;
    x6= x5 + x5;
    x8= x7 + x7;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = abs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double i_iti_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  int x1, x2, x3, x4, x5, x6, x7, x8;

  x1=x2=x3=x4=x5=x6=x7=x8 = 19;

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 * x2;
    x3= x4 * x4;
    x5= x6 * x6;
    x7= x8 * x8;
    x2= x1 * x1;
    x4= x3 * x3;
    x6= x5 * x5;
    x8= x7 * x7;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = abs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double dpp_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  double x1, x2, x3, x4, x5, x6, x7, x8;

  x1=x2=x3=x4=x5=x6=x7=x8 = exp(1.0);

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1++ ;
    x2++ ;
    x3++ ;
    x4++ ;
    x5++ ;
    x6++ ;
    x7++ ;
    x8++ ;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double d_dpd_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  double x1, x2, x3, x4, x5, x6, x7, x8, c;

  x1=x2=x3=x4=x5=x6=x7=x8= exp(1.0);
  c = -exp(1.0);

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 + x2;
    x3= x4 + x4;
    x5= x6 + x6;
    x7= x8 + x8;
    x2= x1 + c;
    x4= x3 + c;
    x6= x5 + c;
    x8= x7 + c;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;

#if DEBUG
  printf("d_dpd_t: x8 = %g \n", (double) x8);
#endif

  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double d_dtd_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, c, t;

  x1=x2=x3=x4=x5=x6=x7=x8 = exp(1.0);
  c = exp(-1.0);

#if DEBUG
  printf("d_dtd_t: n = %g \n", (double)n);
#endif

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 * x2;
    x3= x4 * x4;
    x5= x6 * x6;
    x7= x8 * x8;
    x2= x1 * c ;
    x4= x3 * c;
    x6= x5 * c;
    x8= x7 * c;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;

#if DEBUG
  printf("d_dtd_t: x8 = %g \n", x8);
#endif

  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double sin_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, t;

  x1=x2=x3=x4=x5=x6=x7=x8 = exp(1.0);

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= sin(x2);
    x3= sin(x4);
    x5= sin(x6);
    x7= sin(x8);
    x2= sin(x1+x1);
    x4= sin(x3+x3);
    x6= sin(x5+x5);
    x8= sin(x7+x7);
  }
  t = (clock() - t) / CLOCKS_PER_SEC;

#if DEBUG
  printf("d_sin_t: x7 = %g, x8 = %g \n", x7, x8);
#endif

  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double cos_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, t;

  x1=x2=x3=x4=x5=x6=x7=x8 = 3;

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= cos(x2);
    x3= cos(x4);
    x5= cos(x6);
    x7= cos(x8);
    x2= cos(x1+x1);
    x4= cos(x3+x3);
    x6= cos(x5+x5);
    x8= cos(x7+x7);
  }

#if DEBUG
  printf("d_cos_t: x7 = %g, x8 = %g \n", x7, x8);
#endif

  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);
  return  t;
}
/************************************************************/

double log_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, t;
  double y1, y2, y3, y4, y5, y6, y7, y8;

  x1=x2=x3=x4=x5=x6=x7=x8 = 40*atan(1.0);

  t  = clock();
  for (i = 0; i != n; ++i) {
    y1= log(x1);
    y2= log(x2);
    y3= log(x3);
    y4= log(x4);
    y5= log(x5);
    y6= log(x6);
    y7= log(x7);
    y8= log(x8);
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  y1 = fabs(y1+y2+y3+y4+y5+y6+y7+y8);
  if (y1 < 0) return t + (i < 1);
  return  t;
}
/************************************************************/

double exp_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, t;
  double y1, y2, y3, y4, y5, y6, y7, y8;

  x1=x2=x3=x4=x5=x6=x7=x8=log(40*atan(1));

  t  = clock();
  for (i = 0; i != n; ++i) {
    y1= exp(x1);
    y2= exp(x2);
    y3= exp(x3);
    y4= exp(x4);
    y5= exp(x5);
    y6= exp(x6);
    y7= exp(x7);
    y8= exp(x8);
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  y1 = fabs(y1+y2+y3+y4+y5+y6+y7+y8);
  if (y1 < 0) return t + (i < 1);
  return  t;
}
/************************************************************/

double sqrt_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double x1, x2, x3, x4, x5, x6, x7, x8, t;
  double y1, y2, y3, y4, y5, y6, y7, y8;

  x1=x2=x3=x4=x5=x6=x7=x8=40*atan(1);

  t  = clock();
  for (i = 0; i != n; ++i) {
    y1= sqrt(x1);
    y2= sqrt(x2);
    y3= sqrt(x3);
    y4= sqrt(x4);
    y5= sqrt(x5);
    y6= sqrt(x6);
    y7= sqrt(x7);
    y8= sqrt(x8);
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  y1 = fabs(y1+y2+y3+y4+y5+y6+y7+y8);
  if (y1 < 0) return t + (i < 1);
  return  t;
}
/************************************************************/

double sprintf_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t, x=fabs(40*atan(1));
  char s1[32], s2[32], s3[32], s4[32], s5[32], s6[32], s7[32], s8[32];

  s1[31]=s2[31]=s3[31]=s4[31]=s5[31]=s6[31]=s7[31]=s8[31]= 0;

  t  = clock();
  for (i = 0; i != n; ++i) {
    sprintf(s1, "x=%g, i=%d\n", x, i);
    sprintf(s2, "x=%g, i=%d\n", x, i);
    sprintf(s3, "x=%g, i=%d\n", x, i);
    sprintf(s4, "x=%g, i=%d\n", x, i);
    sprintf(s5, "x=%g, i=%d\n", x, i);
    sprintf(s6, "x=%g, i=%d\n", x, i);
    sprintf(s7, "x=%g, i=%d\n", x, i);
    sprintf(s8, "x=%g, i=%d\n", x, i);
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  i  = s1[31]+s2[31]+s3[31]+s4[31]+s5[31]+s6[31]+s7[31]+s8[31];
  if (x < 0) return  t + (i < 1) ;
  return  t;
}
/************************************************************/


double fpp_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  float  x1, x2, x3, x4, x5, x6, x7, x8;

  x1=x2=x3=x4=x5=x6=x7=x8=4*atan(1);

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1++ ;
    x2++ ;
    x3++ ;
    x4++ ;
    x5++ ;
    x6++ ;
    x7++ ;
    x8++ ;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return t + (i < 1);
  return  t;
}
/************************************************************/

double f_fpf_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/

  ULONG i;
  double t;
  float  x1, x2, x3, x4, x5, x6, x7, x8, c;

  x1=x2=x3=x4=x5=x6=x7=x8= exp(1.0);
  c = -exp(1.0);

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 + x2;
    x3= x4 + x4;
    x5= x6 + x6;
    x7= x8 + x8;
    x2= x1 + c;
    x4= x3 + c;
    x6= x5 + c;
    x8= x7 + c;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i == 0);

#if DEBUG
  printf("f_fpf_t: x8     = %g \n", (double) x8  );
  printf("f_fpf_t: t      = %g \n",  t   );
#endif

  return  t;
}
/************************************************************/

double f_ftf_t (ULONG n)
{
/*
return time in seconds for 8*n calls
*/
  ULONG i;
  double t;
  float x1, x2, x3, x4, x5, x6, x7, x8, c;

  x1=x2=x3=x4=x5=x6=x7=x8 = exp(1.0);
  c = exp(-1.0);

#if DEBUG
  printf("f_ftf_t: n = %g \n", (double)n);
#endif

  t  = clock();
  for (i = 0; i != n; ++i) {
    x1= x2 * x2;
    x3= x4 * x4;
    x5= x6 * x6;
    x7= x8 * x8;
    x2= x1 * c ;
    x4= x3 * c;
    x6= x5 * c;
    x8= x7 * c;
  }
  t = (clock() - t) / CLOCKS_PER_SEC;
  x1 = fabs(x1+x2+x3+x4+x5+x6+x7+x8);
  if (x1 < 0) return  t + (i < 1);

#if DEBUG
  printf("f_ftf_t: x8     = %g \n", (double) x8  );
  printf("f_ftf_t: t      = %g \n", (double) t   );
#endif

  return  t;
}
